'''
This file contains the API-like elements that are called from the outside
'''

import inStrain.profile.profile_controller

def profile_bam(bam, fasta_db, sR2M, ISP_loc, **kwargs):
    '''
    Profile a bam file with inStrain

    arguments:
        bam = location of .bam file
        Fdb = dictionary listing fasta locations to profile
        sR2M = dictionary of scaffold -> read pair -> number of mm
        ISP_loc = location of inStrain profile to store results
    '''
    return inStrain.profile.profile_controller.BamProfileController(
                                        bam, fasta_db, sR2M, ISP_loc, **kwargs).main()



#!/usr/bin/env python

import os
import sys
import shutil
import argparse
import subprocess
import pandas as pd

from Bio import SeqIO
from collections import defaultdict

import inStrain
import inStrain.genomeUtilities

def main(args):
    '''
    The main controller of the program
    '''
    # Parse and validate arguments
    coverm_loc, genome2length = parse_validate(args)

    # Run coverm
    Cdb = run_coverm(coverm_loc, args)

    # Calculate on the genome level
    CGdb = parse_coverm(Cdb, genome2length, args)

    # Produce output
    produce_output(CGdb, args)

def parse_validate(args):
    '''
    Make sure all is well
    '''
    # Make sure coverM is working
    loc = shutil.which('coverm')
    works = False
    if loc != None:
        try:
            o = subprocess.check_output([loc, '-h'],stderr= subprocess.STDOUT)
            works = True
        except:
            pass
    if not works:
        print("Cannot find coverm; make sure its installed")
        sys.exit()

    # Set up a results folder
    if args.output is None:
        args.output = os.path.basename(args.bam)[:-4]
    try:
        os.mkdir(args.output)
    except:
        print("Couldnt make the output directory {0}".format(args.output))
    if args.output[-1] != '/':
        args.output += '/'

    # Set up the stb
    #args.stb = parse_stb(args.stb)
    args.stb = inStrain.genomeUtilities.load_scaff2bin(args.stb)

    # Get genome to length
    scaff2sequence = SeqIO.to_dict(SeqIO.parse(args.fasta, "fasta"))
    s2l = {s:len(scaff2sequence[s]) for s in list(scaff2sequence.keys())}
    del scaff2sequence

    genome2length = {}
    for scaffold, length in s2l.items():
        if scaffold not in args.stb:
            continue
        genome = args.stb[scaffold]
        if genome not in genome2length:
            genome2length[genome] = 0
        genome2length[genome] += (s2l[scaffold])# - 150) # 150 to account for the ignored ends; but it doesn't do that anymore

    return loc, genome2length

def run_coverm(exe_loc, args):
    '''
    run coverm
    '''
    coverm_cmd = "coverm contig -b {0} -t {1} -m mean covered_bases length count --output-format sparse".format(
                args.bam, args.processes)
    coverm_cmd += " | awk 'NR == 1 {{print $0}} NR != 1 {{if ($4/$5 > {0}) print $0}}'".format(args.stringent_breadth_cutoff)

    coverm_cmd += " > {0}".format(args.output + 'coverm_raw.tsv')

    print(coverm_cmd)
    subprocess.call(coverm_cmd, shell=True)

    Cdb = pd.read_csv(args.output + 'coverm_raw.tsv', sep='\t')
    return Cdb

def parse_coverm(Cdb, genome2length, args):
    '''
    Make this genome-wide
    '''
    # Load the stb file
    stb = args.stb
    assert type(stb) == type({})

    # Apply it
    Cdb['genome'] = Cdb['Contig'].map(stb)

    # Print results
    inS = len(Cdb[~Cdb['genome'].isna()]['Contig'].unique())
    tS = len(Cdb['Contig'].unique())
    print("{0:,} of {1:,} scaffolds are in the .stb ({2:.1f}%)".format(inS, tS, (inS/tS)*100))

    inR = Cdb[~Cdb['genome'].isna()]['Read Count'].sum()
    tR = Cdb['Read Count'].sum()
    print("{0:,} of {1:,} mapping reads are in the .stb ({2:.1f}%)".format(inR, tR, (inR/tR)*100))

    # Calculate breadth, coverage, and total reads of genomes
    table = defaultdict(list)
    for genome, db in Cdb.groupby('genome'):
        genome_length = genome2length[genome]
        covered_bases = db['Covered Bases'].sum() / genome_length
        averaged_coverage = sum([b * l for b, l in zip(db['Mean'], db['Length'])]) / genome_length

        table['genome'].append(genome)
        table['length'].append(genome_length)
        table['breadth'].append(covered_bases)
        table['coverage'].append(averaged_coverage)
        table['reads'].append(db['Read Count'].sum())
    CGdb = pd.DataFrame(table)

    return CGdb

def produce_output(CGdb, args):
    '''
    Make stuff
    '''
    # Save the parsed file
    CGdb.to_csv(args.output + 'genomeCoverage.csv', index=False)

    # Save the scaffolds
    genomes = set(CGdb[CGdb['breadth'] >= args.breadth_cutoff]['genome'].unique())
    with open(args.output + 'scaffolds.txt', 'w') as o:
        for scaffold, bin in args.stb.items():
            if bin in genomes:
                o.write(scaffold + '\n')


# if __name__ == '__main__':
#
#     parser = argparse.ArgumentParser(description= """
#         A script to quickly determine breadth and coverage of a .bam file\n
#         """, formatter_class=argparse.RawTextHelpFormatter)
#
#     # Required positional arguments
#     parser.add_argument('-b', '--bam', help="A bam file to profile",
#                         required=True)
#     parser.add_argument('-f', '--fasta', help="The .fasta file to profile",
#                         required=True)
#     parser.add_argument('-s', '--stb', help="Scaffold to bin file",
#                         required=True)
#     parser.add_argument("-o", "--output", action="store", \
#                         help='Output prefix')
#     parser.add_argument("-p", "--processes", action="store", default=6, type=int, \
#                         help='Threads to use for multiprocessing')
#
#     args = parser.parse_args()
#     main(args)



import os
import sys
import shutil
import uuid
import logging


def generate_working_dir(working_dir_base):
    """
    Creates a unique working directory to combat job multitenancy
    :param working_dir_base: base working directory
    :return: a unique subfolder in working_dir_base with a uuid
    """

    working_dir = os.path.join(working_dir_base, str(uuid.uuid4()))
    try:
        os.makedirs(working_dir, exist_ok=True)
    except Exception as e:
        print(e)
        return working_dir_base
    return working_dir


def delete_working_dir(working_dir):
    """
    Deletes working directory
    :param working_dir:  working directory
    """

    try:
        shutil.rmtree(working_dir)
    except Exception as e:
        print ('Can\'t delete %s' % working_dir)

def setup_logger(loc):
    ''' set up logger such that DEBUG goes only to file, rest go to file and console '''

    # Cancel if a logger already exists:
    if logging.getLogger('').handlers:
        return

    # set up logging everything to file
    logging.basicConfig(level=logging.DEBUG,
                       format='%(asctime)s %(levelname)-8s %(message)s',
                       datefmt='%m-%d %H:%M',
                       filename=loc)

    # set up logging of INFO or higher to sys.stderr
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    formatter = logging.Formatter('%(message)s')
    console.setFormatter(formatter)

    logging.getLogger('').addHandler(console)

    logging.debug("!"*80)
    logging.debug("***Logger started up at {0}***".format(loc))
    logging.debug("Command to run was: {0}\n".format(' '.join(sys.argv)))
    logging.debug("!"*80 + '\n')



#!/usr/bin/env python
"""
Run tests
"""

import warnings

import pandas as pd

import tests.test_compare
import tests.test_filter_reads
import tests.test_genes
import tests.test_genome_level
import tests.test_plotting
import tests.test_profile
import tests.test_quick_profile
import tests.test_snv_profile
import tests.tests_special

warnings.filterwarnings("ignore")

pd.set_option('display.max_columns', None)
pd.set_option('display.max_colwidth', -1)


if __name__ == '__main__':
    # tests.test_profile.test_profile().run(tests=[16])
    # tests.test_compare.test_readcomparer().run(tests=[13])

    tests.test_profile.test_profile().run()
    tests.test_filter_reads.TestFilterReads().run()
    tests.test_snv_profile.test_SNVprofile().run()
    tests.test_genes.test_gene_statistics().run()
    tests.test_quick_profile.test_quickProfile().run()
    tests.test_genome_level.test_genome_wide().run()
    tests.test_plotting.test_plot().run()
    tests.test_compare.test_readcomparer().run()
    tests.tests_special.test_special().run()

    print('everything is working swimmingly!')



import logging
import os
import time
import traceback
from collections import defaultdict
from datetime import datetime

import matplotlib
import numpy as np
import pandas as pd
import psutil

matplotlib.use('Agg')
from matplotlib import pyplot as plt


def process_logs(IS_loc):
    """
    Generate the run report for an IS location
    """
    logging.shutdown()
    logloc = os.path.join(IS_loc, 'log/log.log')
    report_run_stats(logloc, most_recent=False, printToo=True, debug=True, save=True)

def report_run_stats(logloc, save=True, most_recent=True, printToo=True,
                    debug=False, plot=False):
    """
    Overall wrapper of this module.

    1) The text log is parsed into a dataframe with "load_log"
    2) A number of reports are generated with "generate_reports"
    3) These reports are saved / displayed here

    """
    if logloc is None:
        return

    # Load the log
    try:
        Ldb = load_log(logloc, most_recent=most_recent)

    except BaseException as e:
        if debug:
            print('Failed to load log file - {1}'.format('None', str(e)))
            traceback.print_exc()
        return

    # Generate reports
    for run, ldb in Ldb.groupby('run_ID'):
        # Make the multi-processing log
        MLdb = load_multiprocessing_log(ldb)

        # Make the reports
        name2report = generate_reports(ldb, MLdb, debug=debug)

        # Print
        if printToo:
            for name, report in name2report.items():
                print("..:: {0} ::..\n{1}".format(name, report))

        if save == True:
            savelocs = [logloc.replace('log.log', '{0}.runtime_summary.txt'.format(run)),
                        logloc.replace('log.log', 'latest.runtime_summary.txt'.format(run))]
            figloc = logloc.replace('log.log', '{0}.ramProfile.png'.format(run))
        elif save == False:
            continue
        else:
            saveloc = save
            figloc = save + '.mem_usage.png'

        for saveloc in savelocs:
            with open(saveloc, 'w') as o:
                for name, report in name2report.items():
                    o.write("..:: {0} ::..\n{1}\n".format(name, report))

    # # Make the plot
    # if plot:
    #     try:
    #         profile_plot(Ldb, saveloc=figloc)
    #     except BaseException as e:
    #         if debug:
    #             print('Failed to make profile plot - {1}'.format('None', str(e)))
    #             traceback.print_exc()

def load_log(logfile, most_recent=True):
    """
    A bunch of messy logic to turn the text file into a dataframe

    The resulting dataframe has 4 columns:
    * log_type = a description of the information that the log contains
    * time = time (in seconds since epoch) that the log was made
    * parseable_string = a string with arbitrary information
    * run_ID = an identifier of a particular instrain run (time when started)

    There are the following types of logs:
    * program_start
    * program_end
    * checkpoint
        - Made using "log_checkpoint". Used for logging regular checkpoints
    * Special_genes
    * Plotting
    * WorkerLog
    * Failure

    """
    table = defaultdict(list)
    with open(logfile) as o:
        prev_line_2 = None
        prev_line = None
        run_ID = None
        for line in o.readlines():
            line = line.strip()
            linewords = [x.strip() for x in line.split()]

            # load new inStrain run
            if 'inStrain version' in line:
                epoch_time = log_fmt_to_epoch("{0} {1}".format(linewords[0], linewords[1]))

                run_ID = datetime.fromtimestamp(epoch_time).strftime('%Y%m%d_%H%M%S')
                cmd = prev_line_2.strip().split('was:')[1].strip()

                table['log_type'].append('program_start')
                table['time'].append(epoch_time)
                table['parsable_string'].append("version={0}; cmd={1}".format(linewords[5], cmd))
                table['run_ID'].append(run_ID)

            # load inStrain run finish
            elif 'inStrain complete' in line:
                epoch_time = log_fmt_to_epoch("{0} {1}".format(linewords[0], linewords[1]))

                table['log_type'].append('program_end')
                table['time'].append(epoch_time)
                table['parsable_string'].append("loglog={0}".format(linewords[14][:-1]))
                table['run_ID'].append(run_ID)

            # regular checkpoints
            elif (len(linewords) > 3) and ('Checkpoint' == linewords[3]):
                epoch_time = log_fmt_to_epoch("{0} {1}".format(linewords[0], linewords[1]))

                table['log_type'].append('checkpoint')
                table['time'].append(epoch_time)
                table['run_ID'].append(run_ID)
                table['parsable_string'].append("class={0};name={1};status={2};RAM={3}".format(
                        linewords[4], linewords[5], linewords[6], linewords[7]))

            # Special gene multiprocessing reporting
            elif 'SpecialPoint_genes' in line:
                pstring = "scaffold={0};PID={1};status={2};what={3}".format(
                            linewords[1], linewords[3], linewords[5], linewords[4])

                table['log_type'].append('Special_genes')
                table['time'].append(float(linewords[6]))
                table['parsable_string'].append(pstring)
                table['run_ID'].append(run_ID)

            elif "Plotting plot" in line:
                epoch_time = log_fmt_to_epoch("{0} {1}".format(linewords[0], linewords[1]))
                pstring = "plot={0}".format(
                            linewords[5])

                table['log_type'].append('Plotting')
                table['time'].append(epoch_time)
                table['parsable_string'].append(pstring)
                table['run_ID'].append(run_ID)

            # Profile reporting
            elif (len(linewords) > 3) and ('WorkerLog' == linewords[0]):
                pstring = "unit={0};PID={1};status={2};process_RAM={3};command={4}".format(
                        linewords[2], linewords[6], linewords[3], linewords[4], linewords[1])

                table['log_type'].append(linewords[0])
                table['time'].append(float(linewords[5]))
                table['parsable_string'].append(pstring)
                table['run_ID'].append(run_ID)

            # Special Failure
            elif 'FAILURE' in line:
                epoch_time = log_fmt_to_epoch("{0} {1}".format(linewords[0], linewords[1]))
                fail_type = linewords[4]

                if fail_type == 'FilterReads':
                    pstring = "type={0};scaffold={1}".format(fail_type, linewords[5])

                elif fail_type == 'SplitException':
                    pstring = "type={0};scaffold={1};split={2}".format(fail_type, linewords[5], linewords[6])

                elif fail_type == 'MergeError':
                    pstring = "type={0};scaffold={1}".format(fail_type, linewords[5])

                elif fail_type == 'GeneException':
                    pstring = "type={0};scaffold={1}".format(fail_type, linewords[5])

                elif fail_type == 'StbError':
                    pstring = "type={0};scaffold={1};bin={2};stb_type={3}".format(
                                fail_type, linewords[5], linewords[6], linewords[7])

                elif fail_type == 'iRepError':
                    pstring = "type={0};genome={1};mm={2}".format(
                                fail_type, linewords[5], linewords[6])

                else:
                    pstring = "type={0}".format(fail_type)

                table['log_type'].append('Failure')
                table['time'].append(float(epoch_time))
                table['parsable_string'].append(pstring)
                table['run_ID'].append(run_ID)

            # Failture that needs to be captured better
            elif 'Double failure!' in line:
                epoch_time = log_fmt_to_epoch("{0} {1}".format(linewords[0], linewords[1]))

                table['log_type'].append('Failure')
                table['time'].append(epoch_time)
                table['parsable_string'].append("error={0}".format(line.strip()))
                table['run_ID'].append(run_ID)

            prev_line_2 = prev_line
            prev_line = line

    Ldb = pd.DataFrame(table)

    # Filter the log
    if most_recent:
        Ldb = filter_most_recent(Ldb)
        assert len(Ldb['run_ID'].unique()) == 1

    return Ldb

def generate_reports(Ldb, MLdb, debug=False):
    """
    Using the parsed log dataframe, make "name2report".

    The following reports are made:
    * Overall
    * Checkpoints
    (lots more; a lot of this is redundent)

    """
    name2report = {}
    assert len(Ldb['run_ID'].unique()) == 1

    # Make the overall report
    name = 'Overall'
    try:
        report, OVERALL_RUNTIME = gen_overall_report(Ldb)
        name2report[name] = report
    except BaseException as e:
        if debug:
            print('Failed to make log for {0} - {1}'.format(name, str(e)))
            traceback.print_exc()

    # Make the main checkpoint report
    name = 'Checkpoints'
    try:
        report = gen_checkpoint_report(Ldb, overall_runtime=OVERALL_RUNTIME.total_seconds(), log_class='main_profile')
        name2report[name] = report
    except BaseException as e:
        if debug:
            print('Failed to make log for {0} - {1}'.format(name, str(e)))
            traceback.print_exc()

    # Make the filter reads reaport
    name = "Filter reads report"
    try:
        # Checkpoint log
        report = ''
        report += gen_checkpoint_report(Ldb, log_class='FilterReads')

        # Set up multiprocessing log
        report += '\n'
        report += gen_multiprocessing_report(MLdb, commands=['GetPairedReads'])
        name2report[name] = report
    except BaseException as e:
        if debug:
            print('Failed to make log for {0} - {1}'.format(name, str(e)))
            traceback.print_exc()

    # Make the profile RAM report
    name = 'Profile report'
    try:
        # Set up checkpoint log
        report = ''
        report += gen_checkpoint_report(Ldb, log_class='Profile')

        # Set up multiprocessing log
        report += '\n* Profiling splits *\n'
        report += gen_multiprocessing_report(MLdb, commands=['SplitProfile'])
        report += '\n* Merging splits and profiling genes *\n'
        report += gen_multiprocessing_report(MLdb, commands=['MergeProfile'])

        name2report[name] = report
    except BaseException as e:
        if debug:
            print('Failed to make log for {0} - {1}'.format(name, str(e)))
            traceback.print_exc()

    # # Make the genes report
    # name = 'Genes paralellization efficiency'
    # try:
    #     report = _gen_genes_report(Ldb)
    #     name2report[name] = report
    # except BaseException as e:
    #     if debug:
    #         print('Failed to make log for {0} - {1}'.format(name, str(e)))
    #         traceback.print_exc()

    # Make the genes report
    name = 'Geneome level report'
    try:
        report = ''
        report += gen_checkpoint_report(Ldb, log_class='GenomeLevel')
        name2report[name] = report
    except BaseException as e:
        if debug:
            print('Failed to make log for {0} - {1}'.format(name, str(e)))
            traceback.print_exc()

    name = 'Plotting'
    try:
        report = _gen_plotting_report(Ldb)
        name2report[name] = report
    except BaseException as e:
        if debug:
            print('Failed to make log for {0} - {1}'.format(name, str(e)))
            traceback.print_exc()

    name = 'Compare'
    try:
        # Checkpoint log
        report = ''
        report += gen_checkpoint_report(Ldb, log_class='Compare')

        # Set up multiprocessing log
        report += '\n'
        report += gen_multiprocessing_report(MLdb, commands=['Compare'])

        name2report[name] = report
    except BaseException as e:
        if debug:
            print('Failed to make log for {0} - {1}'.format(name, str(e)))
            traceback.print_exc()

    # Make the failure report
    name = 'Failures'
    try:
        report = _gen_failures_report(Ldb)
        name2report[name] = report
    except BaseException as e:
        if debug:
            print('Failed to make log for {0} - {1}'.format(name, str(e)))
            traceback.print_exc()

    return name2report

def gen_overall_report(Ldb):
    """
    Report on the overall runtime of the program
    """
    report = ''
    if 'program_end' in Ldb['log_type'].tolist():
        start = datetime.fromtimestamp(Ldb[Ldb['log_type'] == 'program_start']['time'].tolist()[0])
        end = datetime.fromtimestamp(Ldb[Ldb['log_type'] == 'program_end']['time'].tolist()[0])
        runtime = end - start

        start = start.strftime('%Y-%m-%d %H:%M:%S')
        end = end.strftime('%Y-%m-%d %H:%M:%S')
        n2a = parse_parsable_string(Ldb[Ldb['log_type'] == 'program_start']['parsable_string'].tolist()[0])

        report += 'InStrain version {3} started at {0} and ended at {1}.\nRuntime = {2}\n'.format(
                    start, end, td_format(runtime), n2a['version'])
        report += 'Command = {0}\n'.format(n2a['cmd'])
    else:
        start = datetime.fromtimestamp(Ldb[Ldb['log_type'] == 'program_start']['time'].tolist()[0])
        end = datetime.fromtimestamp(Ldb['time'].max())
        runtime = end - start

        start = start.strftime('%Y-%m-%d %H:%M:%S')
        end = end.strftime('%Y-%m-%d %H:%M:%S')
        n2a = parse_parsable_string(Ldb[Ldb['log_type'] == 'program_start']['parsable_string'].tolist()[0])

        report += 'InStrain version {3} started at {0} and ended at {1}.\nCOMMAND FAILED AND DID NOT FINISH.\nRuntime = {2}\n'.format(
                    start, end, td_format(runtime), n2a['version'])
        report += 'Command = {0}\n'.format(n2a['cmd'])

    OVERALL_RUNTIME = runtime
    return report, OVERALL_RUNTIME

def load_multiprocessing_log(Ldb):
    """
    Generate a log table specifically for "WorkerLog" types
    """
    # Parse the initial datatable
    ldb = Ldb[(Ldb['log_type'] == 'WorkerLog')]
    table = defaultdict(list)
    for i, row in ldb.iterrows():
        for thing, value in parse_parsable_string(row['parsable_string']).items():
            table[thing].append(value)
        table['time'].append(row['time'])
    Ldb = pd.DataFrame(table)
    return Ldb

def gen_multiprocessing_report(MLdb, commands):
    """
    From MLdb
    """
    # Load multiprocessing
    if len(MLdb) > 0:
        mldb = MLdb[MLdb['command'].isin(commands)]
        if len(mldb) > 0:
            ldb = parse_multiprocessing(mldb)
            return gen_multiprocessing_text(ldb)

    return ''

def parse_multiprocessing(Ldb):
    """
    Ldb looks like:

                              unit    PID status  process_RAM  command  \
    0    N5_271_010G1_scaffold_132  32031  end    188485632.0  Compare
    1    N5_271_010G1_scaffold_75   32032  end    188555264.0  Compare
    2    N5_271_010G1_scaffold_34   32033  end    188387328.0  Compare
    3    N5_271_010G1_scaffold_152  32035  end    187920384.0  Compare
    4    N5_271_010G1_scaffold_135  32036  end    188030976.0  Compare
    ..                         ...    ...  ...            ...      ...
    173  N5_271_010G1_scaffold_96   32035  end    198172672.0  Compare
    174  N5_271_010G1_scaffold_15   32033  end    201838592.0  Compare
    175  N5_271_010G1_scaffold_32   32034  end    199385088.0  Compare
    176  N5_271_010G1_scaffold_10   32031  end    198950912.0  Compare
    177  N5_271_010G1_scaffold_8    32036  end    199233536.0  Compare

                 time
    0    1.597446e+09
    1    1.597446e+09
    2    1.597446e+09
    3    1.597446e+09
    4    1.597446e+09


    """
    table = defaultdict(list)
    Ldb['time'] = Ldb['time'].astype(float)
    Ldb['process_RAM'] = Ldb['process_RAM'].astype(float)
    first_time = Ldb['time'].min()

    # Generate this on a per-unit level
    for scaffold, ddb in Ldb.groupby('unit'):
        for cmd, db in ddb.groupby('command'):
            sdb = db[db['status'] == 'start']
            edb = db[db['status'] == 'end']

            table['unit'].append(scaffold)
            table['PID'].append(db['PID'].iloc[0])

            table['start_time'].append(sdb['time'].iloc[0])
            table['adjusted_start'].append(sdb['time'].iloc[0] - first_time)
            table['start_process_RAM'].append(sdb['process_RAM'].iloc[0])

            if len(edb) > 0:
                table['adjusted_end'].append(edb['time'].iloc[0] - first_time)
                table['end_process_RAM'].append(edb['process_RAM'].iloc[0])
                table['end_time'].append(edb['time'].iloc[0])
            else:
                for i in ['adjusted_end', 'end_process_RAM', 'end_time']:
                    table[i].append(np.nan)

            table['runs'].append(len(sdb))
            table['command'].append(cmd)

    db = pd.DataFrame(table)
    db['runtime'] = [s - e for s, e in zip(db['end_time'], db['start_time'])]
    db['RAM_usage'] = [s - e for s, e in zip(db['end_process_RAM'], db['start_process_RAM'])]

    return db

def gen_multiprocessing_text(rdb, name='unit'):
    report = ''

    # Overall wall time
    start = datetime.fromtimestamp(rdb['start_time'].min())
    end = datetime.fromtimestamp(rdb['end_time'].max())
    runtime = end - start

    # User time
    parallel_time = rdb['runtime'].sum()
    avg_time = rdb['runtime'].mean()

    # Number of processes used
    PIDs = len(rdb['PID'].unique())

    report += "{0:30}\t{1}\n".format("Wall time", td_format(runtime))
    report += "{0:30}\t{1}\n".format("Total processes used", PIDs)
    report += "{0:30}\t{1:.1f}\n".format("Average number processes used", parallel_time/runtime.total_seconds())
    report += "{0:30}\t{1:.1f}%\n".format("Paralellization efficiency", (parallel_time/runtime.total_seconds()/(PIDs))*100)
    report += "{0:30}\t{1}\n".format("Units profiled", len(rdb['unit'].unique()))

    # Report on splits
    #report += "\n"
    report += "{0:30}\t{1}\n".format("Average time per unit", td_format(None, seconds=rdb['runtime'].mean()))
    report += "{0:30}\t{1}\n".format("Average time per unit", td_format(None, seconds=rdb['runtime'].mean()))
    report += "{0:30}\t{1}\n".format("Median time per unit", td_format(None, seconds=rdb['runtime'].median()))
    report += "{0:30}\t{1}\n".format("Maximum unit time", td_format(None, seconds=rdb['runtime'].max()))
    report += "{0:30}\t{1}\n".format("Longest running unit", rdb.sort_values('runtime', ascending=False)['unit'].iloc[0])
    report += "{0:30}\t{1}\n".format("Per-process efficiency", sorted(["{0:.1f}".format((d['runtime'].sum()/(rdb['end_time'].max() - d['start_time'].min()))*100) for p, d in rdb.groupby('PID')]))

    # Report on RAM
    #report += "\n"
    report += "{0:35}\t{1}\n".format("{0} per-process strating RAM".format(name),
                ["{0}".format(humanbytes(d['start_process_RAM'].iloc[0])) for p, d in rdb.groupby('PID')])
    report += "{0:35}\t{1}\n".format("{0} per-process final RAM".format(name),
                ["{0}".format(humanbytes(d['start_process_RAM'].iloc[-1])) for p, d in rdb.groupby('PID')])
    report += "{0:35}\t{1}\n".format("{0} per-process minimum RAM".format(name),
                ["{0}".format(humanbytes(d['start_process_RAM'].min())) for p, d in rdb.groupby('PID')])
    report += "{0:35}\t{1}\n".format("{0} per-process maximum RAM".format(name),
                ["{0}".format(humanbytes(d['start_process_RAM'].max())) for p, d in rdb.groupby('PID')])

    return report

def gen_checkpoint_report(ldb, overall_runtime=None, log_class=None):
    """
    The prefered way of handling checkpoints

    ldb should have the columns:
        name = name of checkpoint
        status = start or end
        RAM = RAM usage
        time = time in epoch time
    """
    # Parse the parseable strings
    ldb = ldb[ldb['log_type'] == 'checkpoint']
    if len(ldb) > 0:
        for i in ['name', 'class', 'status', 'RAM']:
            ldb[i] = [parse_parsable_string(pstring)[i] for pstring in ldb['parsable_string']]
    else:
        return ''

    # Subset to a specific class
    if log_class is not None:
        ldb = ldb[ldb['class'] == log_class]
    if len(ldb) == 0:
        return ''

    # Handle the overall runtime
    if overall_runtime is None:
        overall_runtime = max(ldb[ldb['status'] == 'end']['time'].max() - ldb[ldb['status'] == 'start']['time'].min(), 1)

    report = ''
    order = list(ldb['name'].unique())

    for name in order:
        db = ldb[ldb['name'] == name]
        if len(db) > 2:
            report += '{0} has problems and cannot be reported\n'.format(name)

        elif len(db) == 2:
            start = db[db['status'] == 'start'].iloc[0]['time']
            end = db[db['status'] == 'end'].iloc[0]['time']
            runtime = end - start

            if 'RAM' in ldb.columns:
                startram = int(db[db['status'] == 'start']['RAM'].tolist()[0])
                endram = int(db[db['status'] == 'end']['RAM'].tolist()[0])
                ram_change = endram-startram

                if ram_change > 0:
                    inc_dec = 'increased'
                else:
                    inc_dec = 'decreased'

                report += '{0:20} took {1:15} ({2:4.1f}% of overall)\tRAM went from {5} to {6} ({4} by {3})\n'.format(
                            name, td_format(None, seconds=runtime), (runtime/overall_runtime)*100,
                            humanbytes(ram_change, sign=False), inc_dec,
                            humanbytes(startram, sign=False),
                            humanbytes(endram, sign=False))

            else:
                report += '{0:20} took {1:15} ({2:4.1f}% of overall)\n'.format(name, td_format(runtime), (runtime/overall_runtime)*100)

        elif len(db) == 1:
            try:
                start = db[db['status'] == 'start'].iloc[0]['time']
                report += '{0:20} started at {1} and never finished\n'.format(name, start)
            except:
                report += '{0:20} failed\n'.format(name)

    return report

def _gen_genes_report(Ldb, detailed=False):
    report = ''

    # Set up checkpoint log
    report += gen_checkpoint_report(Ldb, log_class='GeneProfile')
    if report != '':
        report += '\n'

    # Set up paralellization log
    ldb = Ldb[Ldb['log_type'] == 'Special_genes']
    PGdb = _load_genes_logtable(ldb)

    if len(PGdb) == 0:
        return report

    # Generate report on paralellization as a whole
    rdb = PGdb[PGdb['command'] == 'whole']

    start = datetime.fromtimestamp(rdb['start_time'].min())
    end = datetime.fromtimestamp(rdb['end_time'].max())
    runtime = end - start
    parallel_time = rdb['runtime'].sum()
    avg_time = rdb['runtime'].mean()
    PIDs = len(rdb['PID'].unique())

    # Report on paralelized steps as a whole
    report += "{0:30}\t{1}\n".format("Wall time parallelized steps", td_format(runtime))
    report += "{0:30}\t{1}\n".format("Total processes used", PIDs)
    report += "{0:30}\t{1:.1f}\n".format("Average number processes used", parallel_time/runtime.total_seconds())
    report += "{0:30}\t{1:.1f}%\n".format("Paralellization efficiency", (parallel_time/runtime.total_seconds()/(PIDs))*100)
    report += "{0:30}\t{1}\n".format("Scaffolds profiled", len(rdb['scaffold'].unique()))
    report += "{0:30}\t{1}\n".format("Average time per scaffold", td_format(None, seconds=rdb['runtime'].mean()))
    report += "{0:30}\t{1}\n".format("Median time per scaffold", td_format(None, seconds=rdb['runtime'].median()))
    report += "{0:30}\t{1}\n".format("Maximum split scaffold", td_format(None, seconds=rdb['runtime'].max()))
    report += "{0:30}\t{1}\n".format("Longest running scaffold", rdb.sort_values('runtime', ascending=False)['scaffold'].iloc[0])
    report += "{0:30}\t{1}\n".format("Per-process efficiency", sorted(["{0:.1f}".format((d['runtime'].sum()/(rdb['end_time'].max() - d['start_time'].min()))*100) for p, d in rdb.groupby('PID')]))

    # Report on sub-paralalized steps
    for step, db in PGdb.groupby('command'):
        if step == 'whole':
            continue

        step_start = datetime.fromtimestamp(db['start_time'].min())
        step_start = datetime.fromtimestamp(db['end_time'].max())
        step_parallel_time = db['runtime'].sum()

        report += "{0:30}\t{1:.1f}% of parallel runtime\n".format("Step {0}: ".format(step), (step_parallel_time/parallel_time)*100)

    return report

def _gen_failures_report(Ldb):
    report = ''
    ldb = Ldb[Ldb['log_type'] == 'Failure']
    if len(ldb) > 0:
        ldb['failure_type'] = [parse_parsable_string(p)['type'] for p in ldb['parsable_string']]

        for t, db in ldb.groupby('failure_type'):
            table = defaultdict(list)
            for i, row in db.iterrows():
                for thing, value in parse_parsable_string(row['parsable_string']).items():
                    table[thing].append(value)
                table['time'].append(row['time'])
            fdb = pd.DataFrame(table)

            if t == 'FilterReads':
                report += "The following scaffolds were not in the bam file:\n"
                for s in fdb['scaffold'].unique():
                    report += s + '\n'
                report += '\n'

            elif t == 'SplitException':
                report += "The following splits failed during profiling:\n"
                for i, row in fdb.iterrows():
                    report += "{0} split {1}\n".format(row['scaffold'], row['split'])
                report += '\n'

            elif t == 'MergeError':
                report += "The following scaffolds could not be profiled due to mering errors:\n"
                for i, row in fdb.iterrows():
                    report += "{0}\n".format(row['scaffold'])
                report += '\n'

            elif t == 'GeneException':
                report += "Genes on the following scaffolds could not be profiled due to errors durring profiling:\n"
                for i, row in fdb.iterrows():
                    report += "{0}\n".format(row['scaffold'])
                report += '\n'

            elif t == 'StbError':
                report += "The following scaffolds were in the .stb file given, but not the original .fasta file " \
                        + "used for profiling. They will not be considered in genomeLevel operations:\n"
                for i, row in fdb.iterrows():
                    report += "{0} (intended for genome {1})\n".format(row['scaffold'], row['bin'])
                report += '\n'

            elif t == 'iRepError':
                report += "The following genomes failed to calculate iRep for an unknown reason:\n"
                for i, row in fdb.iterrows():
                    report += "{0} (mm {1})\n".format(row['genome'], row['mm'])
                report += '\n'

            else:
                report += "I dont know how to report {0} failures\n".format(t)
                for i, row in fdb.iterrows():
                    report += str(row) + '\n'
                report += '\n'

    if report == '':
        report = "No failures"
    return report

def _gen_plotting_report(Ldb):
    report = ''
    ldb = Ldb[Ldb['log_type'] == 'Plotting']
    if len(ldb) > 0:
        ldb['plot'] = [parse_parsable_string(p)['plot'] for p in ldb['parsable_string']]
        ldb = ldb.sort_values('time').reset_index(drop=True)
        for i, row in ldb.iterrows():
            if row['plot'] == 'finished':
                break

            start = row['time']
            end = ldb.iloc[i+1]['time']
            report += "Plot {0} took {1}\n".format(row['plot'], td_format(None, seconds=end-start))

    return report

def _load_genes_logtable(ldb):
    table = defaultdict(list)
    for i, row in ldb.iterrows():
        for thing, value in parse_parsable_string(row['parsable_string']).items():
            table[thing].append(value)
        table['time'].append(row['time'])
    Ldb = pd.DataFrame(table)

    if len(Ldb) == 0:
        return Ldb

    table = defaultdict(list)
    Ldb['time'] = Ldb['time'].astype(float)
    first_time = Ldb['time'].min()
    for scaffold, ddb in Ldb.groupby('scaffold'):
        for cmd, db in ddb.groupby('what'):
            sdb = db[db['status'] == 'start']
            edb = db[db['status'] == 'end']

            table['scaffold'].append(scaffold)
            table['PID'].append(db['PID'].tolist()[0])
            table['start_time'].append(sdb['time'].tolist()[0])
            table['end_time'].append(edb['time'].tolist()[0])
            table['adjusted_start'].append(sdb['time'].tolist()[0] - first_time)
            table['adjusted_end'].append(edb['time'].tolist()[0] - first_time)
            table['runs'].append(len(sdb))
            table['command'].append(cmd)

    db = pd.DataFrame(table)
    db['runtime'] = [s-e for s,e in zip(db['end_time'], db['start_time'])]

    return db

def td_format(td_object, seconds=None):
    if seconds is None:
        seconds = int(td_object.total_seconds())
    periods = [
        ('year',        60*60*24*365),
        ('month',       60*60*24*30),
        ('day',         60*60*24),
        ('hour',        60*60),
        ('minute',      60),
        ('second',      1)
    ]

    strings=[]
    for period_name, period_seconds in periods:
        if seconds > period_seconds:
            period_value , seconds = divmod(seconds, period_seconds)
            has_s = 's' if period_value > 1 else ''
            strings.append("%s %s%s" % (period_value, period_name, has_s))

    if len(strings) > 0:
        return ", ".join(strings)
    else:
        return "<1 second"

def humanbytes(B, sign=True):
    if B < 0:
        if sign:
            unit = '- '
        else:
            unit = ''
        B = B * -1
    else:
        unit = ''

    B = float(B)
    KB = float(1024)
    MB = float(KB ** 2) # 1,048,576
    GB = float(KB ** 3) # 1,073,741,824
    TB = float(KB ** 4) # 1,099,511,627,776

    if B < KB:
        return '{2}{0} {1}'.format(B,'Bytes' if 0 == B > 1 else 'Byte', unit)
    elif KB <= B < MB:
        return '{1}{0:.2f} KB'.format(B/KB, unit)
    elif MB <= B < GB:
        return '{1}{0:.2f} MB'.format(B/MB, unit)
    elif GB <= B < TB:
        return '{1}{0:.2f} GB'.format(B/GB, unit)
    elif TB <= B:
        return '{1}{0:.2f} TB'.format(B/TB, unit)

def parse_parsable_string(pstring):
    object2string = {}
    linewords = pstring.split(';')
    for word in linewords:
        ws = word.split('=')
        object2string[ws[0].strip()] = ws[1].strip()
    return object2string

def filter_most_recent(Ldb):
    """
    Only keep the most recent run
    """
    ID = Ldb.sort_values('time')['run_ID'].tolist()[-1]
    return Ldb[Ldb['run_ID'] == ID]

def log_fmt_to_epoch(ttime):
    # Old log format with no year
    if len(ttime.split('-')) == 2:
        oldformat = '%m-%d %H:%M'
        datetimeobject = datetime.strptime(ttime,oldformat)
        datetimeobject = datetimeobject.replace(year=datetime.now().year)
    # New log format with year
    else:
        oldformat = '%y-%m-%d %H:%M:%S'
        datetimeobject = datetime.strptime(ttime,oldformat)

    return datetimeobject.timestamp()

def log_checkpoint(log_class, name, status, inc_children=True):
    """
    Log a checkpoint for the program in the debug under "log" status

    Arguments:
        log_class    = where is this log comming from ("main_profile", "GeneProfile", etc.)
        name         = what is the task that you're logging ("load globals", "run loop", etc.)
        stauts       = either the string "start" and "end"
        inc_children = include child processes as well in RAM tally

    Results:
        Make the following log message:
        "Checkpoint class task status RAM"

        When split on tabs, this has the following structure:
        "Checkpoint" = linewords[3]
        class = linewords[4]
        task = linewords[5]
        start/end = linewords[6]
        RAM = linewords[7]
    """
    current_process = psutil.Process(os.getpid())
    mem = current_process.memory_info().rss
    if inc_children:
        for child in current_process.children(recursive=True):
            try:
                mem += child.memory_info().rss
            except:
                pass

    assert status in ['start', 'end'], [log_class, name, status]
    assert len(name.split()) == 1

    logging.debug("Checkpoint {0} {1} {2} {3}".format(log_class, name, status, mem))

def get_worker_log(worker_type, unit, status, inc_children=False):
    """
    Return a string with log information intended to be generated within a worker process

    Arguments:
        worker_type  = The type of worker this is
        unit         = The unit this worker is currently processing
        status       = start / end
        inc_children = include child processes as well in RAM tally

    Returns:
        A string with the following structure:
        "WorkerLog worker_type unit status RAM time PID"

        "WorkerLog" = linewords[0]
        worker_type = linewords[1]
        unit = linewords[2]
        status = linewords[3]
        ram = linewords[4]
        time = linewords[5]
        PID = linewords[6]

    """
    pid = os.getpid()
    current_process = psutil.Process(pid)
    mem = current_process.memory_info().rss
    if inc_children:
        for child in current_process.children(recursive=True):
            try:
                mem += child.memory_info().rss
            except:
                pass

    assert status in ['start', 'end'], status

    return "\nWorkerLog {0} {1} {2} {3} {4} {5}".format(worker_type, unit, status, mem, time.time(), pid)


# Get the version
from ._version import __version__

# Import packages
import gc
import os
import sys
import h5py
import copy
import pysam
import logging
import argparse
import pandas as pd

from subprocess import call
from datetime import datetime
from collections import defaultdict

# Import inStrain stuff
import inStrain.profile
import inStrain.profile.samtools_ops

import inStrain.filter_reads
import inStrain.readComparer
import inStrain.GeneProfile
import inStrain.genomeUtilities
import inStrain.plottingUtilities
import inStrain.quickProfile
import inStrain.SNVprofile
import inStrain.logUtils

class Controller():
    '''
    Controller of the whole shebang
    '''

    def main(self, args):
        ''' Parse user options and call the correct pipeline'''
        # Call the appropriate workflow
        if args.operation == "profile":
            self.profile_operation(args)

        if args.operation == "compare":
            self.compare_operation(args)

        if args.operation == "filter_reads":
            self.filter_reads_operation(args)

        if args.operation == "profile_genes":
            self.profile_genes_operation(args)

        if args.operation == "genome_wide":
            self.genome_wide_operation(args)

        if args.operation == "quick_profile":
            self.quick_profile_operation(args)

        if args.operation == "plot":
            self.plot_operation(args)

        if args.operation == "other":
            self.other_operation(args)

        self.shutdown(args)

    def profile_operation(self, args):
        ProfileController(args).main()

    def compare_operation(self, args):
        inStrain.readComparer.main(args)

    def filter_reads_operation(self, args):
        inStrain.filter_reads.Controller().main(args)

    def profile_genes_operation(self, args):
        inStrain.GeneProfile.Controller().main(args)

    def genome_wide_operation(self, args):
        inStrain.genomeUtilities.Controller().main(args)

    def quick_profile_operation(self, args):
        inStrain.quickProfile.main(args)

    def plot_operation(self, args):
        inStrain.plottingUtilities.main(args)

    def other_operation(self, args):
        # Check if you should convert IS profile
        if args.old_IS != None:
            inStrain.SNVprofile.convert_SNVprofile(args.old_IS)
        if args.run_statistics != None:
            inStrain.logUtils.process_logs(args.run_statistics)

    def shutdown(self, args):
        try:
            logloc = logging.getLoggerClass().root.handlers[0].baseFilename
        except:
            return
        logging.debug("inStrain complete; shutting down logger and printing run stats (log location = {0})".format(logloc))
        logging.shutdown()
        inStrain.logUtils.report_run_stats(logloc, most_recent=True, printToo=args.debug, debug=args.debug)

class ProfileController(object):
    '''
    Main controller of the profile command
    '''
    def __init__(self, args):
        '''
        Set all of the command line arguments in the "args" attribute

        Doing it this way lets your pass the arguments to other controllers
        '''
        self.args = args
        self.ori_args = copy.deepcopy(args)

    def main(self):
        '''
        The main method when run on the command line
        '''
        # Parse arguments
        self.validate_arguments()

        # Filter reads
        self.profile_filter_reads()

        # Profile
        self.run_profile()

        # Profile genes
        # self.profile_profile_genes()

        # Make things genome_wide
        self.profile_genome_wide()

        # Make plots
        self.profile_plots()

        # Final message
        self.write_final_message()

        return self.ISP

    def validate_arguments(self):
        '''
        Do some parsing, start up a logger
        '''
        # Get out the "args" to manipulate it
        args = self.args

        # default prefix is now fasta prefix -alexcc 5/8/2019
        if args.output == 'inStrain':
            args.output = args.fasta.split(".")[0].split("/")[-1]

        # Set up "base"
        out_base = args.output

        # Set up Logger
        outbase = out_base
        ISP = inStrain.SNVprofile.SNVprofile(outbase)
        log_loc = ISP.get_location('log') + 'log.log'
        setup_logger(log_loc)

        # Make the bam file if you need to; remove it from args
        self.bam = inStrain.profile.samtools_ops.prepare_bam_fie(args.bam, args.processes)
        del self.args.bam

        # Load the list of scaffolds
        args.scaffolds_to_profile = inStrain.profile.fasta.load_scaff_list(
                                                    args.scaffolds_to_profile)

        # Fix the fdr
        if args.fdr == 0:
            args.fdr = 1e-6

        # Handle database mode
        if args.database_mode:
            args.min_read_ani = 0.92
            args.skip_mm_profiling = True
            args.min_genome_coverage = 1

        # Make sure you have a .stb if needed
        if args.min_genome_coverage != 0:
            assert args.stb != [], 'If you adjust the minimum genome coverage, you need to provide an .stb!'

        # See if you have genes
        if args.gene_file != None:
            self.genes_file_present = True

        self.args = args
        self.args.IS = ISP.get('location')
        self.ISP = ISP
        self.kwargs = vars(self.args)

    def profile_filter_reads(self):
        '''
        Call the filter reads module as run with "profile"
        '''
        message = """\
***************************************************
    ..:: inStrain profile Step 1. Filter reads ::..
***************************************************
        """
        logging.info(message)
        inStrain.logUtils.log_checkpoint("main_profile", "filter_reads", "start")

        # Profile reads
        Rdic, RR, fasta_db, scaff2sequence, s2l = \
                inStrain.filter_reads.Controller().main_from_profile(self.ISP,
                                                    self.bam, **self.kwargs)

        # Store results
        self.RR = RR
        self.Rdic = Rdic
        self.fasta_db = fasta_db
        self.scaff2sequence = scaff2sequence
        self.scaffold2length = s2l

        # Parse results
        inStrain.logUtils.log_checkpoint("FilterReads", "parse_results", "start")
        self.parse_filter_reads()
        inStrain.logUtils.log_checkpoint("FilterReads", "parse_results", "end")

        inStrain.logUtils.log_checkpoint("main_profile", "filter_reads", "end")

    def parse_filter_reads(self):
        '''
        Parse filter read results in the context of profile
        '''
        # Parse results
        self.scaffold2pairs = self.RR.set_index('scaffold')['filtered_pairs'].to_dict()
        self.readlength = float(self.RR.loc[0, 'mean_pair_length'])

        # Filter the .fasta file with these results
        self.fasta_db = inStrain.profile.fasta.filter_fasta(self.fasta_db,
                        self.scaffold2pairs, self.scaffold2length,
                        self.readlength, **self.kwargs)

        # Change up Rdic if needed
        if self.args.skip_mm_profiling:
            newRdic = {}
            for s, p2i in self.Rdic.items():
                newRdic[s] = set(p2i.keys())
            self.Rdic = newRdic
            self.ISP.store('Rdic', self.Rdic, 'pickle', 'list of filtered read pairs')
        else:
            self.ISP.store('Rdic', self.Rdic, 'dictionary', 'Read pair -> mismatches')

        # Handle exceptions
        if self.RR['filtered_pairs'].tolist()[0] == 0:
            logging.error("Because no read pairs remain I'm going to crash now. Maybe this is failing because you dont have paired reads (in which case you should adjust --pairing_filter option), or maybe its failing because the mapper you used uses full fasta headers (in which case you should use the flag --use_full_fasta_header)")
            raise Exception('No paired reads detected; see above message and log')

        if len(self.fasta_db) <= 0:
            logging.error("No scaffolds passed initial filtering based on numbers of mapped reads")
            raise Exception('No scaffolds detected; see above message and log')

        # Store results
        self.ISP.store('mapping_info', self.RR, 'pandas', "Report on reads")
        self.ISP.store('fasta_loc', os.path.abspath(self.args.fasta), 'value', 'Location of .fasta file used during profile')
        self.ISP.store('scaffold2length', self.scaffold2length, 'dictionary', 'Dictionary of scaffold 2 length')

        # Print status report
        unfiltered_pairs = self.RR['unfiltered_pairs'].iloc[0]
        filterd_pairs = self.RR['filtered_pairs'].iloc[0]
        pair_length = self.readlength

        status = ''
        status += "{0:.1f}% of reads were removed during filtering\n".format(
                    ((unfiltered_pairs - filterd_pairs) / unfiltered_pairs)*100)
        status += "{0:,} read pairs remain ({1:#.4g} Gbp)".format(
                    filterd_pairs, (filterd_pairs * pair_length)/1e9)
        logging.info(status)

    def run_profile(self):
        '''
        Call the actual profile module
        '''
        message = """\
***************************************************
.:: inStrain profile Step 2. Profile scaffolds ::..
***************************************************
        """
        logging.info(message)
        inStrain.logUtils.log_checkpoint("main_profile", "profile_scaffolds", "start")

        # Do some argument handling
        self.kwargs['s2s'] = self.scaff2sequence
        self.kwargs['s2p'] = self.scaffold2pairs

        # Call the module
        self.ISP = inStrain.profile.profile_bam(self.bam, self.fasta_db,
                                                self.Rdic, self.ISP.get('location'),
                                                **self.kwargs)

        # Write output
        inStrain.logUtils.log_checkpoint("Profile", "store_output", "start")
        self.write_output()
        inStrain.logUtils.log_checkpoint("Profile", "store_output", "end")

        inStrain.logUtils.log_checkpoint("main_profile", "profile_scaffolds", "end")

    def write_output(self):
        '''
        Write output files
        '''
        logging.debug("Writing output files now")

        for t in ['SNVs', 'scaffold_info', 'SNVs', 'linkage', 'gene_info']:
            self.ISP.generate(t)
        self.ISP.generate('mapping_info', **self.kwargs)

    def profile_profile_genes(self):
        '''
        Call profile genes from the "profile" module
        '''
        message = """\
***************************************************
  .:: inStrain profile Step 3. Profile genes ::..
***************************************************
        """
        logging.info(message)
        args = self.args

        if args.gene_file != None:
            inStrain.logUtils.log_checkpoint("main_profile", "profile_genes", "start")
            Controller().profile_genes_operation(copy.deepcopy(args))
            inStrain.logUtils.log_checkpoint("main_profile", "profile_genes", "end")
        else:
            logging.info('Nevermind! You didnt include a genes file')

    def profile_genome_wide(self):
        '''
        Call genome_wide from "profile" module
        '''
        message = """\
***************************************************
.:: inStrain profile Step 4. Make genome-wide ::..
***************************************************
        """
        logging.info(message)
        args = self.args

        if not args.skip_genome_wide:
            inStrain.logUtils.log_checkpoint("main_profile", "genome_wide", "start")
            Controller().genome_wide_operation(copy.deepcopy(args))
            inStrain.logUtils.log_checkpoint("main_profile", "genome_wide", "end")
        else:
            logging.info('Nevermind! You chose to skip genome_wide')

    def profile_plots(self):
        '''
        Call plotting function from "profile" module
        '''
        # Generate plots
        message = """\
***************************************************
 .:: inStrain profile Step 5. Generate plots ::..
***************************************************
        """
        logging.info(message)
        args = self.args

        if not args.skip_plot_generation:
            inStrain.logUtils.log_checkpoint("main_profile", "making_plots", "start")
            args.plots = 'a'
            Controller().plot_operation(args)
            inStrain.logUtils.log_checkpoint("main_profile", "making_plots", "end")
        else:
            logging.info('Nevermind! You chose to skip making plots')

    def write_final_message(self):
        Sprofile = self.ISP
        message = """\
$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

..:: inStrain profile finished ::..

Output tables........ {0}
Figures.............. {1}
Logging.............. {2}

See documentation for output descriptions - https://instrain.readthedocs.io/en/latest/

$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
        """.format(Sprofile.get_location('output'), \
            Sprofile.get_location('figures'),
            Sprofile.get_location('log'))
        logging.info(message)

def setup_logger(loc):
    ''' set up logger such that DEBUG goes only to file, rest go to file and console '''

    # Cancel if a logger already exists:
    if logging.getLogger('').handlers:
        return

    # set up logging everything to file
    logging.basicConfig(level=logging.DEBUG,
                       format='%(asctime)s %(levelname)-8s %(message)s',
                       datefmt='%y-%m-%d %H:%M:%S',
                       filename=loc)

    # set up logging of INFO or higher to sys.stderr
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    formatter = logging.Formatter('%(message)s')
    console.setFormatter(formatter)

    logging.getLogger('').addHandler(console)

    logging.debug("!"*80)
    logging.debug("***Logger started up at {0}***".format(loc))
    logging.debug("Command to run inStrain was: {0}\n".format(' '.join(sys.argv)))
    logging.debug("inStrain version {0} was run \n".format(__version__))
    logging.debug("!"*80 + '\n')



#!/usr/bin/env python

def humanbytes(B):
   '''
   Return the given bytes as a human friendly KB, MB, GB, or TB string
   https://stackoverflow.com/questions/12523586/python-format-size-application-converting-b-to-kb-mb-gb-tb/37423778
   '''
   B = float(B)
   KB = float(1024)
   MB = float(KB ** 2) # 1,048,576
   GB = float(KB ** 3) # 1,073,741,824
   TB = float(KB ** 4) # 1,099,511,627,776

   if B < KB:
      return '{0} {1}'.format(B,'Bytes' if 0 == B > 1 else 'Byte')
   elif KB <= B < MB:
      return '{0:.2f} KB'.format(B/KB)
   elif MB <= B < GB:
      return '{0:.2f} MB'.format(B/MB)
   elif GB <= B < TB:
      return '{0:.2f} GB'.format(B/GB)
   elif TB <= B:
      return '{0:.2f} TB'.format(B/TB)



'''
Handles code related to linkage
'''

import itertools
import collections
import numpy as np
import pandas as pd
import networkx as nx
from collections import defaultdict

import inStrain.profile.profile_utilities

def calc_mm_SNV_linkage_network(read_to_snvs, scaff=False):
    '''
    Calculates the SNV linkage network

    Arguments:
        read_to_snvs = mm -> read -> SNVs

    saves it as a dictionary of edges in self.snv_net.
    Writes it out to a file, genome.net for reading in through other programs like node2vec
    '''

    G = nx.Graph() # graph between positions. has the attribute mm2combo2counts
    for mm, tread2snvs in read_to_snvs.items():

        for read in tread2snvs:
            snvs = tread2snvs[read]
            for pair in itertools.combinations(snvs, 2):
                p1 = int(pair[0].split(":")[0])
                p2 = int(pair[1].split(":")[0])

                c = "{0}:{1}".format(pair[0].split(":")[1], pair[1].split(":")[1])
                if not G.has_edge(p1, p2):
                    G.add_edge(p1, p2)
                    G[p1][p2]['mm2combo2counts'] = {}
                if mm not in G[p1][p2]['mm2combo2counts']:
                    G[p1][p2]['mm2combo2counts'][mm] = {}
                if c not in G[p1][p2]['mm2combo2counts'][mm]:
                    G[p1][p2]['mm2combo2counts'][mm][c] = 0
                G[p1][p2]['mm2combo2counts'][mm][c] += 1

    return G

def calculate_ld(mm_to_position_graph, min_snp,
                snv2mm2counts, scaffold):
    '''
    Calculates Linkage Disequilibrium for all SNVs in a window.
    '''
    r2_table = defaultdict(list)

    # This just gives potisions
    for edge in mm_to_position_graph.edges(list(snv2mm2counts.keys())):

        # For this position, now you have combo (allele combo) to mm (read mismatches) to counts (reads wih that combo and level of mismatches)
        mm2combo2counts = mm_to_position_graph.edges[edge[0],edge[1]]['mm2combo2counts']
        # Proceed with the full
        for mm, ld_args in _iterator_ld_sites(mm2combo2counts, min_snp, snv2mm2counts,
                    p1=int(edge[0]),p2=int(edge[1])):

            ld_result = _calc_ld_single(min_snp, **ld_args)

            if ld_result is False:
                continue

            ld_result['distance'] = abs(int(edge[0]) - int(edge[1]))
            ld_result['position_A'] = int(edge[0])
            ld_result['position_B'] = int(edge[1])

            _update_r2(ld_result, r2_table, mm, scaffold)

    # Create r2 linkage table
    r2linkage_table = pd.DataFrame(r2_table)
    return r2linkage_table

P2C = {'A':0, 'C':1, 'T':2, 'G':3}
def _iterator_ld_sites(mm2combo2counts, min_snp, snv2mm2counts, p1, p2):
    '''
    pass
    '''
    # Get the tables ready
    s1 = snv2mm2counts[p1]
    s2 = snv2mm2counts[p2]

    # Figure out when you need to update
    updateMMs = set(s1.keys()).intersection(set(s2.keys()))

    counts = {0:np.zeros(4), 1:np.zeros(4)}
    combocounts = defaultdict(int)


    for mm, combo2counts in sorted(mm2combo2counts.items()):
        # update combo counts
        for combo, count in combo2counts.items():
            combocounts[combo] += count

        # update counts
        if mm not in updateMMs:
            continue
        counts[0] = inStrain.profile.profile_utilities.mm_counts_to_counts(s1, mm)
        counts[1] = inStrain.profile.profile_utilities.mm_counts_to_counts(s2, mm)

        if sum(counts[0]) + sum(counts[1]) >= min_snp:
            skip = False
            arg_dic = {}

            arg_dic['A'], arg_dic['a'] = major_minor_allele(counts[0])
            #assert arg_dic['A'] != arg_dic['a']

            arg_dic['B'], arg_dic['b'] = major_minor_allele(counts[1])
            #assert arg_dic['B'] != arg_dic['b']

            for val, key in zip(['A', 'a', 'B', 'b'], [0, 0, 1, 1]):
                count = int(counts[key][P2C[arg_dic[val]]])
                arg_dic[val + '_counts'] = count
                if count == 0:
                    skip = True

            if skip:
                continue

            total = 0
            for a in ['A', 'a']:
                for b in ['B', 'b']:
                    val = combocounts[arg_dic[a] + ':' + arg_dic[b]]
                    arg_dic[a + b] = val
                    total += val
            arg_dic['total'] = total

            yield mm, arg_dic

def major_minor_allele(counts):
    d = {'A': counts[0], 'C': counts[1], 'T': counts[2], 'G': counts[3]}
    l = sorted(d, key=d.get, reverse=True)
    return l[0], l[1]

def _calc_ld_single(min_snp, **kwargs):
    '''
    A function that calculates the LD between two SNPs

    Argumements:
        snp_a = position of SNP A
        snp_b = position of SNP B

    Notes:
        allele_A = major allele at position A
        allele_a = minor allele at position A
    '''

    #distance between snps

    # figure out alleles
    allele_A = kwargs.get('A')
    allele_a = kwargs.get('a')
    allele_B = kwargs.get('B')
    allele_b = kwargs.get('b')
    countAB = kwargs.get('AB', 0)
    countAb = kwargs.get('Ab', 0)
    countaB = kwargs.get('aB', 0)
    countab = kwargs.get('ab', 0)

    total = kwargs.get('total')

    #Requires at least min_snp
    if total > min_snp:

        freq_AB = float(countAB) / total
        freq_Ab = float(countAb) / total
        freq_aB = float(countaB) / total
        freq_ab = float(countab) / total

        freq_A = freq_AB + freq_Ab
        freq_a = freq_ab + freq_aB
        freq_B = freq_AB + freq_aB
        freq_b = freq_ab + freq_Ab

        linkD = freq_AB - freq_A * freq_B

        if freq_a == 0 or freq_A == 0 or freq_B == 0 or freq_b == 0:
            r2 = np.nan
        else:
            r2 = linkD*linkD / (freq_A * freq_a * freq_B * freq_b)

        linkd = freq_ab - freq_a * freq_b

        # calc D-prime
        d_prime = np.nan
        if (linkd < 0):
            denom = max([(-freq_A*freq_B),(-freq_a*freq_b)])
            d_prime = linkd / denom

        elif (linkD > 0):
            denom = min([(freq_A*freq_b), (freq_a*freq_B)])
            d_prime = linkd / denom

        ################
        # calc rarefied

        rareify = np.random.choice(['AB','Ab','aB','ab'], replace=True, p=[freq_AB,freq_Ab,freq_aB,freq_ab], size=min_snp)
        freq_AB = float(collections.Counter(rareify)['AB']) / min_snp
        freq_Ab = float(collections.Counter(rareify)['Ab']) / min_snp
        freq_aB = float(collections.Counter(rareify)['aB']) / min_snp
        freq_ab = float(collections.Counter(rareify)['ab']) / min_snp

        freq_A = freq_AB + freq_Ab
        freq_a = freq_ab + freq_aB
        freq_B = freq_AB + freq_aB
        freq_b = freq_ab + freq_Ab

        linkd_norm = freq_ab - freq_a * freq_b

        if freq_a == 0 or freq_A == 0 or freq_B == 0 or freq_b == 0:
            r2_normalized = np.nan
        else:
            r2_normalized = linkd_norm*linkd_norm / (freq_A * freq_a * freq_B * freq_b)



        # calc D-prime
        d_prime_normalized = np.nan
        if (linkd_norm < 0):
            denom = max([(-freq_A*freq_B),(-freq_a*freq_b)])
            d_prime_normalized = linkd_norm / denom

        elif (linkd_norm > 0):
            denom = min([(freq_A*freq_b), (freq_a*freq_B)])
            d_prime_normalized = linkd_norm / denom

        rt_dict = {}
        for att in ['r2', 'd_prime', 'r2_normalized', 'd_prime_normalized', 'total', 'countAB', \
                    'countAb', 'countaB', 'countab', 'allele_A', 'allele_a', \
                    'allele_B', 'allele_b']:
            rt_dict[att] = eval(att)


        return rt_dict

    else:
        return False

def _update_r2(ld_result, r2_table, mm, scaffold, window=False):
    if ld_result == False:
        return r2_table

    for key, value in ld_result.items():
        r2_table[key].append(value)
    r2_table['mm'].append(mm)
    r2_table['scaffold'].append(scaffold)
    #r2_table['window'].append("{0}:{1}_{2}".format(scaffold, window[0], window[1]))

    return r2_table

def update_linked_reads(read_to_snvs, pileupcolumn, MMcounts, position,
                        bases, R2M, min_freq=0.05, scaffold=False):
    '''
    Find and update linked reads

    try:
        table[R2M[pileupread.alignment.query_name]]\
        [P2C[pileupread.alignment.query_sequence[pileupread.query_position]]] += 1
    '''

    position = str(position)

    # Get reads to snvs
    for pileupread in pileupcolumn.pileups:
        read_name = pileupread.alignment.query_name
        if not pileupread.is_del and not pileupread.is_refskip and read_name in R2M:
            if type(R2M) == type({}):
                mm = R2M[read_name]
            else:
                mm = 0

            try:
                val = pileupread.alignment.query_sequence[pileupread.query_position]

                # if value is not the consensus value
                if val in bases:
                    # this read is of a variant position
                    read_to_snvs[mm][read_name].append(position + ":" + val)
            except KeyError: # This would be like an N or something not A/C/T/G
                pass



#!/usr/bin/env python

import os
import sys
import glob
import shlex
import shutil
import logging
import subprocess
from argparse import ArgumentParser

from s3_utils import download_file, upload_file, download_folder, upload_folder
from job_utils import generate_working_dir, delete_working_dir#, setup_logger


def download_reference(s3_path, working_dir):
    """
    Downloads reference folder that has been configured to run with Isaac
    :param s3_path: S3 path that the folder resides in
    :param working_dir: working directory
    :return: local path to the folder containing the reference
    """

    reference_folder = os.path.join(working_dir, 'reference')

    try:
        os.mkdir(reference_folder)
    except Exception as e:
        pass

    download_folder(s3_path, reference_folder)

    # Update sorted reference
    update_sorted_reference(reference_folder)

    return reference_folder


def download_fastq_files(fastq1_s3_path, fastq2_s3_path, working_dir):
    """
    Downlodas the fastq files
    :param fastq1_s3_path: S3 path containing FASTQ with read1
    :param fastq2_s3_path: S3 path containing FASTQ with read2
    :param working_dir: working directory
    :return: local path to the folder containing the fastq
    """
    fastq_folder = os.path.join(working_dir, 'fastq')

    try:
        os.mkdir(fastq_folder)
    except Exception as e:
        pass

    local_fastq1_path = download_file(fastq1_s3_path, fastq_folder)
    local_fastq2_path = download_file(fastq2_s3_path, fastq_folder)

    # Isaac requires the fastqs to be symlinked as lane1_read1.fastq.gz and lane1_read2.fastq.gz
    os.symlink(local_fastq1_path, os.path.join(fastq_folder, 'lane1_read1.fastq.gz'))
    os.symlink(local_fastq2_path, os.path.join(fastq_folder, 'lane1_read2.fastq.gz'))

    return fastq_folder


def upload_bam(bam_s3_path, local_folder_path):
    """
    Uploads results folder containing the bam file (and associated output)
    :param bam_s3_path: S3 path to upload the alignment results to
    :param local_folder_path: local path containing the alignment results
    """

    upload_folder(bam_s3_path, local_folder_path)


def run_isaac(reference_dir, fastq_folder_path, memory, cmd_args, working_dir):
    """
    Runs Isaac3
    :param reference_dir: local path to directory containing reference
    :param fastq_folder_path: local path to directory containing fastq files
    :param memory: memory in GB
    :param cmd_args: additional command-line arguments to pass in
    :param working_dir: working directory
    :return: path to results
    """

    # Maps to Isaac's folder structure and change working directory
    os.chdir(working_dir)
    bam_folder = os.path.join(working_dir, 'Projects/default/default')

    try:
        os.mkdir(bam_folder)
    except Exception as e:
        pass

    cmd = 'isaac-align -r %s -b %s --base-calls-format fastq-gz -o %s -m %s %s' % \
          (os.path.join(reference_dir, 'sorted-reference.xml'), fastq_folder_path, working_dir, memory, cmd_args)
    print ("Running: %s" % cmd)
    subprocess.check_call(shlex.split(cmd))

    return bam_folder


def update_sorted_reference(reference_dir):
    """
    Updates sorted-reference.xml to map to the correct directory path. Since each analysis occurs in subfolder of the
    working directory, it will change each execution
    :param reference_dir: Reference directory
    """
    with open(os.path.join(reference_dir, 'sorted-reference.xml'), 'r') as infile:
        sorted_reference = infile.read()

    with open(os.path.join(reference_dir, 'sorted-reference.xml'), 'w') as outfile:
        outfile.write(sorted_reference.replace('/scratch', reference_dir))

def download_data(args, working_dir, tmp_dir):
    cmd_string = ''

    # Download the .bam
    if args.bam != None:
        logging.info("Downloading bam to {0}".format(tmp_dir))
        download_file(args.bam, tmp_dir)
        cmd_string = cmd_string + os.path.join(tmp_dir, os.path.basename(args.bam)) + ' '

        try:
            logging.info("Downloading bam index".format(tmp_dir))
            download_file(args.bam + '.bai', tmp_dir)
        except:
            pass

    # Downlaod the .fasta
    if args.fasta != None:
        logging.info("Downloading fasta to {0}".format(tmp_dir))
        download_file(args.fasta, tmp_dir)
        cmd_string = cmd_string + os.path.join(tmp_dir, os.path.basename(args.fasta)) + ' '

    # Download other files
    for f, name in zip([args.genes, args.stb], ['-g', '-s']):
        if f is not None:
            logging.info("{0} is {1}; downloading".format(name, f))
            download_file(f, tmp_dir)
            cmd_string = cmd_string + name + ' ' + os.path.join(tmp_dir, os.path.basename(f)) + ' '

    # 2) Unzip if need be
    to_unzip = glob.glob(tmp_dir + '/*.gz')
    if len(to_unzip) > 0:
        for g in to_unzip:
            cmd = 'gzip -d {0}'.format(g)
            subprocess.check_call(shlex.split(cmd))

    # Get the work directory
    wd_loc = os.path.join(working_dir, args.wd_name)
    cmd_string = cmd_string + ' -o ' + wd_loc

    return cmd_string, wd_loc

def download_checkm(arags):
    pass

def run_inStrain(data_prefix, command, cmd_args, timeout=0):
    '''
    Actually call inStrain

    Args:
        genomes_string: string with genomes, INCLUDING -g argument
        wd_loc: location of output work directory
        command: the string to put right after "command"
        cmd_args: things to put after the command
    '''
    if timeout == 0:
        cmd = 'inStrain {0} {1} {2}'.format(command, data_prefix, cmd_args)
    else:
        cmd = 'timeout {3} inStrain {0} {1} {2}'.format(command, data_prefix, cmd_args, timeout)
    logging.info ("Running: %s" % cmd)
    subprocess.check_call(shlex.split(cmd))

def upload_results(args, working_dir, wd_loc):
    if args.light_upload:
        # delete the raw_data folder
        data_folder = str(glob.glob(wd_loc + '/raw_data/')[0])
        logging.info("deleting {0}".format(data_folder))
        shutil.rmtree(data_folder)

    upload_folder(args.results_directory, working_dir)

def parse_arguments():
    argparser = ArgumentParser()

    file_path_group = argparser.add_argument_group(title='File paths')
    file_path_group.add_argument('--bam', type=str, help='s3 path to the .bam file')
    file_path_group.add_argument('--fasta', type=str, help='s3 path to the .fasta file')
    file_path_group.add_argument('--genes', type=str, help='s3 path to the genes file')
    file_path_group.add_argument('--stb', help='s3 path to the stb file', nargs='*', default=[]))
    file_path_group.add_argument('--results_directory', type=str, help='s3 path to the folder to put the results in')

    run_group = argparser.add_argument_group(title='Run command args')
    run_group.add_argument('--wd_name', type=str, help='Name of the output directory', default='instrain_output')
    run_group.add_argument('--timeout', type=int, help='Kill job after this many seconds and upload what was gotten so far. Input 0 here to have no limit', default=0)
    run_group.add_argument('--command', type=str, help='The command that should go after the command inStrain', default='profile')
    run_group.add_argument('--cmd_args', type=str, help='A string (as long as you want) that will be put after the inStrain command, .bam, fasta, and output directory', default=' ')
    run_group.add_argument('--light_upload', help='By default it will upload the /raw_data folder; this will make it not', default=False, action='store_true')

    argparser.add_argument('--working_dir', type=str, default='/mnt/scratch')

    args = argparser.parse_args()
    return args


def main():
    args = parse_arguments()

    # Make a working directory to be uploaded at the end
    working_dir = generate_working_dir(args.working_dir)
    log_loc = os.path.join(working_dir, 'docker_log.log')
    setup_logger(log_loc)
    logging.info('Here are the args: {0}'.format(args))

    # Get a temporary directory for scratch things
    temp_dir = generate_working_dir('/mnt/temp')

    # Download the required things
    logging.info('downloading data')
    data_prefix, wd_loc = download_data(args, working_dir, temp_dir)

    logging.info('running inStrain command')
    try:
        run_inStrain(data_prefix, args.command, args.cmd_args, args.timeout)
    except:
        pass

    logging.info('uploading results')
    upload_results(args, working_dir, wd_loc)

    logging.info('cleaning up')
    delete_working_dir(working_dir)
    delete_working_dir(temp_dir)

### TO BE MOVED TO COMMON UTILS
def setup_logger(loc):
    ''' set up logger such that DEBUG goes only to file, rest go to file and console '''

    # Cancel if a logger already exists:
    if logging.getLogger('').handlers:
        return

    # set up logging everything to file
    logging.basicConfig(level=logging.DEBUG,
                       format='%(asctime)s %(levelname)-8s %(message)s',
                       datefmt='%m-%d %H:%M',
                       filename=loc)

    # set up logging of INFO or higher to sys.stderr
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    formatter = logging.Formatter('%(message)s')
    console.setFormatter(formatter)

    logging.getLogger('').addHandler(console)

    logging.debug("!"*80)
    logging.debug("***Logger started up at {0}***".format(loc))
    logging.debug("Command to run was: {0}\n".format(' '.join(sys.argv)))
    logging.debug("!"*80 + '\n')
###

if __name__ == '__main__':
    main()



import io
import os
import subprocess
import shlex
import boto3

s3 = boto3.resource('s3')

def download_folder(s3_path, directory_to_download):
    """
    Downloads a folder from s3
    :param s3_path: s3 folder path
    :param directory_to_download: path to download the directory to
    :return: directory that was downloaded
    """
    cmd = 'aws s3 cp --recursive %s %s' % (s3_path, directory_to_download)
    print(cmd)

    subprocess.check_call(shlex.split(cmd))

    return directory_to_download


def download_file(s3_path, directory_to_download):
    """
    Downloads an object from s3 to a local path
    :param s3_path: s3 object path
    :param directory_to_download: directory to download to
    :return: local file path of the object
    """
    bucket = s3_path.split('/')[2]
    key = '/'.join(s3_path.split('/')[3:])

    object_name = key.split('/')[-1]

    local_file_name = os.path.join(directory_to_download, object_name)

    s3.Object(bucket, key).download_file(local_file_name)

    return local_file_name


def upload_folder(s3_path, local_folder_path, sse=False):
    """
    Uploads a local folder to S3
    :param s3_path: s3 path to upload folder to
    :param local_folder_path: local folder path
    :param sse: boolean whether to enable server-side encryption
    """
    cmd = 'aws s3 cp --recursive %s %s' % (local_folder_path, s3_path)

    if sse:
        cmd += ' --sse'

    subprocess.check_call(shlex.split(cmd))


def upload_file(s3_path, local_path):
    """
    Uploads a local file to s3 with server side encryption enabled
    :param s3_path: s3 object path
    :param local_path: local file path
    :return: response from the upload file
    """
    bucket = s3_path.split('/')[2]
    key = '/'.join(s3_path.split('/')[3:])

    response = s3.Object(bucket, key).upload_file(local_path, ExtraArgs=dict(ServerSideEncryption='AES256'))

    return response


def get_matching_s3_objects(bucket, prefix="", suffix=""):
    """
    Generate objects in an S3 bucket.

    :param bucket: Name of the S3 bucket.
    :param prefix: Only fetch objects whose key starts with
        this prefix (optional).
    :param suffix: Only fetch objects whose keys end with
        this suffix (optional).
    """
    s3 = boto3.client('s3')
    paginator = s3.get_paginator("list_objects_v2")

    kwargs = {'Bucket': bucket}

    # We can pass the prefix directly to the S3 API.  If the user has passed
    # a tuple or list of prefixes, we go through them one by one.
    if isinstance(prefix, str):
        prefixes = (prefix, )
    else:
        prefixes = prefix

    for key_prefix in prefixes:
        kwargs["Prefix"] = key_prefix

        for page in paginator.paginate(**kwargs):
            try:
                contents = page["Contents"]
            except KeyError:
                return

            for obj in contents:
                key = obj["Key"]
                if key.endswith(suffix):
                    yield obj

def get_matching_s3_keys(bucket, prefix="", suffix=""):
    """
    Generate the keys in an S3 bucket.

    :param bucket: Name of the S3 bucket.
    :param prefix: Only fetch keys that start with this prefix (optional).
    :param suffix: Only fetch keys that end with this suffix (optional).
    """
    for obj in get_matching_s3_objects(bucket, prefix, suffix):
        yield obj["Key"]

def check_s3_file(floc):
    '''
    Return True if exists and False if it does not
    '''
    bucket = floc.split('/')[2]
    prefix = '/'.join(floc.split('/')[3:])

    found = False
    for key in get_matching_s3_keys(bucket, prefix):
        if prefix in key:
            found = True
    return found

def store_s3_file(bucket, location, binary_string):
    s3 = boto3.resource('s3')
    object = s3.Object(bucket, location)
    object.put(Body=binary_string)


def load_coverage_report(s3_bucket, s3_key, sep='\t', names=None):
    '''
    https://towardsdatascience.com/web-scraping-html-tables-with-python-c9baba21059
    '''
    # Load the data from s3
    client = boto3.client("s3")
    obj = client.get_object(Bucket=s3_bucket, Key=s3_key)
    df = pd.read_csv(io.BytesIO(obj['Body'].read()), sep=sep, names=names)

    return df

def glob_s3(s3_path):
    bucket = s3_path.split('/')[2]
    key = '/'.join(s3_path.split('/')[3:])

    prefix = key.split('*')[0]
    suffix = key.split('*')[1]

    return ['s3://' + bucket + '/' + x for x in get_matching_s3_keys(bucket, prefix, suffix)]
